"""
Path constants used in other modules.

Author: Benedikt Vollmerhaus
License: MIT
"""

import pathlib
import tempfile

#: The cache directory to save transition frames in
CACHE_DIR = pathlib.Path.home() / '.cache/blurwal'

#: The temporary directory to save comparison frames in
TEMP_DIR = pathlib.Path(tempfile.gettempdir()) / 'blurwal'

#: The flat file for storing the original wallpaper's path
ORIGINAL_PATH = CACHE_DIR / 'original-path'


def get_original() -> pathlib.Path:
    """
    Return the original wallpaper's path from the path storage file.

    :return: The original wallpaper's path
    """
    with ORIGINAL_PATH.open() as file:
        return pathlib.Path(file.readline().strip())


def set_original(path: pathlib.Path) -> None:
    """
    Save the given original wallpaper's path in the path storage file.

    :param path: The path to save
    :return: None
    """
    ORIGINAL_PATH.write_text(str(path))



"""
Window and workspace operations using Xlib.

Author: Benedikt Vollmerhaus
License: MIT
"""

import logging
from typing import List

from Xlib import error

from blurwal import x11


def count_open(ignored_classes: List[str], x_conn: x11.XConnection) -> int:
    """
    Count the number of open windows on the focused workspace.

    Windows with a class in the given ignored list, bad windows,
    or ones missing a workspace number property are not counted.

    :param ignored_classes: A list of window classes to ignore
    :param x_conn: A connection to X for window retrieval
    :return: The number of open windows on the workspace
    """
    window_count = 0

    for window in x_conn.get_windows_on_current_workspace():
        try:
            window_class = window.get_wm_class()
        except error.BadWindow:
            logging.info('Ignoring bad window (id: %s)', window.id)
            continue

        if window_class is not None and window_class[1] in ignored_classes:
            logging.info("Ignoring window with class '%s'.", window_class[1])
            continue

        window_count += 1

    return window_count



#!/usr/bin/env python3

"""
Main entry point with CLI argument parsing and startup.

Author: Benedikt Vollmerhaus
License: MIT
"""

import argparse
import logging
import sys
from typing import List

from blurwal import environment, paths, utils, wallpaper
from blurwal._version import __version__
from blurwal.blur import Blur


def parse_args(arg_list: List) -> argparse.Namespace:
    """
    Parse and return any provided command line arguments.

    :return: A namespace holding the parsed arguments
    """
    parser = argparse.ArgumentParser(
        description='Smoothly blurs the wallpaper when windows are opened.',
        epilog="The transition's speed is based on the chosen number of "
               "steps in combination with how quickly the backend can "
               "update the wallpaper on your hardware. This cannot be "
               "changed, as the backend's performance is the limiting "
               "upper factor. To vary the transition speed you should "
               "thus change the number of steps (-s).")

    parser.add_argument('-v', '--version',
                        action='version', version=f'%(prog)s {__version__}')

    parser.add_argument('-m', '--min',
                        type=int, metavar='N', default=2,
                        help='the minimum number of windows to blur the '
                             'wallpaper (default: %(default)d)')

    parser.add_argument('-s', '--steps',
                        type=int, metavar='N', default=10,
                        help='the number of steps in a blur transition; '
                             'see below (default: %(default)d, min: 2)')

    parser.add_argument('-b', '--blur',
                        type=int, metavar='N', default=10,
                        help='the blur strength (sigma) to use when '
                             'fully blurred (default: %(default)d)')

    parser.add_argument('-i', '--ignore',
                        nargs='*', metavar='class', default=[],
                        help='a space-separated list of window classes '
                             'to exclude when counting the number of '
                             'open windows')

    parser.add_argument('--backend',
                        type=str, metavar='name',
                        choices=environment.BACKENDS,
                        help='the wallpaper backend to enable; supported '
                             'ones are: [%(choices)s]')

    parser.add_argument('--verbose',
                        action='store_true',
                        help='print additional information')

    parser.add_argument('--debug',
                        action='store_true',
                        help='print detailed debug output')

    args = parser.parse_args(arg_list)

    if args.steps < 2:
        parser.error('The transition must have at least 2 steps.')

    if args.verbose:
        logging.getLogger().setLevel(logging.INFO)

    if args.debug:
        logging.getLogger().setLevel(logging.DEBUG)

    return args


def main() -> None:
    """
    Retrieve the backend to use, set or restore the original
    wallpaper if necessary, and get ready for fancy blurring!

    :return: None
    """
    logging.basicConfig(format='%(levelname)s: %(message)s')

    args = parse_args(sys.argv[1:])

    backend = environment.get_backend(args.backend)
    if backend is None:
        logging.error('Could not find a compatible backend to use.')
        sys.exit(1)

    environment.prepare(backend)

    current_path = backend.get_current()
    if wallpaper.is_transition(current_path):
        # If the script terminated unexpectedly and a transition
        # frame got stuck as the wallpaper, restore the original.
        backend.restore_original()
    elif current_path is not None:
        utils.print_subheading('Current wallpaper: ')
        utils.print_info(str(current_path))
        paths.set_original(current_path)

    if args.ignore:
        print(f':: Ignoring window classes: {", ".join(args.ignore)}')

    try:
        blur = Blur(args, backend)
        if blur.frames_are_outdated():
            blur.generate_transition_frames()

        blur.listen_for_events()
    except KeyboardInterrupt:
        print('\nBye!')
        sys.exit(0)


if __name__ == '__main__':
    main()



"""
Transition frame generation via ImageMagick's convert utility.

Author: Benedikt Vollmerhaus
License: MIT
"""

import filecmp
import pathlib
import subprocess

from blurwal import paths, utils


def generate(output_dir: pathlib.Path, blur_level: int,
             max_blur_level: int, max_sigma: int) -> None:
    """
    Generate a transition frame by applying a blur to the wallpaper.

    The blur level is given in the range:
      [0, <total no. of transition steps>]

    and then converted to the sigma range:
      [0, <max. sigma when fully blurred>]

    The resulting sigma is then used for ImageMagick's blur operation.
    This means that the highest blur level, i.e. final frame in a blur
    transition, will have been blurred with the given max. sigma value.

    ImageMagick's blur operator needs two values, a radius and a sigma:

    - Radius limits the blurring to pixels that are within that radius
      of the one being blurred. For example, a radius of 1 would limit
      the blurred area to the direct neighbors of each pixel.

    - Sigma determines the blur strength within that radius, i.e. how
      much each of the neighbors in the pixel's radius contributes to
      the blurred area. For example, the largest possible sigma would
      produce a simple averaging of all neighboring pixels.

    See also: https://www.imagemagick.org/Usage/blur/#blur_args

    :param output_dir: Where to save the resulting frame
    :param blur_level: A blur level to blur the wallpaper with
    :param max_blur_level: The max. blur level (total no. of steps)
    :param max_sigma: The sigma to use at the maximum blur level

    :return: None
    """
    output_file = output_dir / f'frame-{blur_level}.jpg'
    sigma = utils.map_range(blur_level, (0, max_blur_level), (0, max_sigma))

    subprocess.run(['convert', paths.get_original(),
                    '-blur', f'0x{sigma}', str(output_file)])


def is_outdated(blur_level: int, max_blur_level: int, max_sigma: int) -> bool:
    """
    Blur the wallpaper with a given blur level and compare the result
    to the already existing frame of that level. If the images differ,
    then the wallpaper changed and the frames need to be regenerated.

    :param blur_level: A blur level whose existing frame to validate
    :param max_blur_level: The max. blur level (total no. of steps)
    :param max_sigma: The sigma to use at the maximum blur level

    :return: Whether the frame of the given blur level is outdated
    """
    generate(paths.TEMP_DIR, blur_level, max_blur_level, max_sigma)

    reference_frame = paths.TEMP_DIR / f'frame-{blur_level}.jpg'
    actual_frame = paths.CACHE_DIR / f'frame-{blur_level}.jpg'

    return not filecmp.cmp(reference_frame, actual_frame)



"""
Author: Benedikt Vollmerhaus
License: MIT
"""

import logging
import pathlib
import re
import sys
from typing import List, Optional

from blurwal.backends import base


class FehBackend(base.Backend):
    """Wallpaper operations utilizing the feh backend."""

    FEHBG_FILE = pathlib.Path.home() / '.fehbg'

    def get_wallpaper_set_cmds(self, path: pathlib.Path) -> List[List[str]]:
        """
        Return a list of commands to be executed to set the wallpaper.

        :param path: The image to set as the wallpaper
        :return: A list of commands to be executed
        """
        return [['feh', '--bg-fill', str(path)]]

    def get_current(self) -> Optional[pathlib.Path]:
        """
        Return the current wallpaper's path from the ~/.fehbg file.

        :return: The current wallpaper's path or None if not retrievable
        """
        try:
            # Search for '-enclosed strings starting with / from back to front
            path = re.search(r"(?s:.*)'(/+.+)'", self.FEHBG_FILE.read_text())
            if not path:
                logging.warning('Could not extract current wallpaper from %s',
                                self.FEHBG_FILE.name)
                return None

            return pathlib.Path(path.group(1))
        except FileNotFoundError:
            logging.exception('Could not open %s', self.FEHBG_FILE.name)
            sys.exit(1)



"""
Author: Benedikt Vollmerhaus
License: MIT
"""

import logging
import threading

from blurwal import paths
from blurwal.backends import base


class Transition(threading.Thread):
    """
    A thread for asynchronously transitioning between two blur levels.

    It supports being stopped to enable reversing a running transition.
    The current blur level is saved on each frame, so that a subsequent
    thread may begin reversing the previous transition from there, even
    if that transition was interrupted.
    """

    def __init__(self, from_blur_level: int, to_blur_level: int,
                 backend: base.Backend):
        super().__init__()
        self._stop_event = threading.Event()

        self._from_blur_level: int = from_blur_level
        self._to_blur_level: int = to_blur_level
        self._backend: base.Backend = backend

        self.current_level: int = from_blur_level

    def stop(self) -> None:
        """
        Stop the currently running blur transition.

        :return: None
        """
        self._stop_event.set()

    def is_stopped(self) -> bool:
        """
        Return whether the blur transition has been stopped.

        :return: Whether the blur transition has been stopped
        """
        return self._stop_event.is_set()

    def run(self) -> None:
        """
        Begin transitioning from the initial to the target blur level
        by changing the wallpaper to each intermediate frame in quick
        succession (speed is dependent on the backend's performance).

        :return: None
        """
        if self._from_blur_level > self._to_blur_level:
            logging.info('Unblurring from blur level %s to %s.',
                         self._from_blur_level, self._to_blur_level)

            blur_levels = reversed(range(self._to_blur_level,
                                         self._from_blur_level))
        else:
            logging.info('Blurring from blur level %s to %s.',
                         self._from_blur_level, self._to_blur_level)

            blur_levels = range(self._from_blur_level + 1,
                                self._to_blur_level + 1)

        for level in blur_levels:
            if self.is_stopped():
                break

            self.current_level = level

            blur_frame = paths.CACHE_DIR / f'frame-{level}.jpg'
            self._backend.change_to(blur_frame)



"""
Functions for preparing the environment to run in.

Author: Benedikt Vollmerhaus
License: MIT
"""

import atexit
import logging
import os
import shutil
from typing import Dict, Optional, Tuple, Type

from blurwal import paths, utils
from blurwal.backends import base, feh, xfconf

#: Supported backend names with their implementation and binary
BACKENDS: Dict[str, Tuple[Type[base.Backend], str]] = {
    'feh': (feh.FehBackend, 'feh'),
    'xfce': (xfconf.XfconfBackend, 'xfconf-query')
}

#: A map of XDG_CURRENT_DESKTOP values to their respective backends
XDG_TO_BACKEND: Dict[str, str] = {'XFCE': 'xfce'}


def get_backend(name: Optional[str]) -> Optional[base.Backend]:
    """
    Return an instance of the wallpaper backend to use.

    The backend is selected with the following priority:
      1. The backend corresponding to the given name, if installed
      2. The backend assigned to the current desktop environment
         (retrieved from the XDG_CURRENT_DESKTOP env variable)
      3. The first installed backend if neither is available

    :param name: The name of the backend to use or None
    :return: A backend instance or None if none is available
    """
    if name:
        if shutil.which(BACKENDS[name][1]):
            utils.print_heading('Using supplied wallpaper backend: ')
            utils.print_info(name)
            return BACKENDS[name][0]()

        logging.error('The given backend is not installed, '
                      'falling back to auto-detection.')

    xdg_desktop: Optional[str] = os.environ.get('XDG_CURRENT_DESKTOP')
    detected_backend: Optional[str] = XDG_TO_BACKEND.get(xdg_desktop, None)
    if detected_backend is None:
        logging.info('XDG_CURRENT_DESKTOP not set or unsupported, '
                     'falling back to first available backend.')

        detected_backend = next((name for name, backend in BACKENDS.items()
                                 if shutil.which(backend[1])), None)
        if detected_backend is None:
            return None

    utils.print_heading('Auto-detected wallpaper backend: ')
    utils.print_info(detected_backend)

    return BACKENDS[detected_backend][0]()


def prepare(backend: base.Backend) -> None:
    """
    Create the required cache/temp directories if not already existing
    and register an exit handler for restoring the original wallpaper.

    :return: None
    """
    paths.CACHE_DIR.mkdir(parents=True, exist_ok=True)
    paths.TEMP_DIR.mkdir(exist_ok=True)

    atexit.register(backend.restore_original)



"""
Author: Benedikt Vollmerhaus
License: MIT
"""

import logging
import pathlib
import subprocess
import sys
from typing import List, Optional

from blurwal import x11
from blurwal.backends import base


def find_wallpaper_properties() -> List[str]:
    """
    Return a list of xfconf-query wallpaper properties to set.

    :return: A list of xfconf-query wallpaper properties to set
    """
    property_list = subprocess.check_output(['xfconf-query',
                                             '-c', 'xfce4-desktop',
                                             '--list'])
    properties = property_list.strip().decode().splitlines()
    return [p for p in properties if p.endswith('last-image')]


class XfconfBackend(base.Backend):
    """Wallpaper operations utilizing Xfce's xfconf-query backend."""

    def __init__(self):
        x_conn = x11.XConnection()
        self._primary_screen = x_conn.get_primary_output()
        self._properties = find_wallpaper_properties()

    def get_wallpaper_set_cmds(self, path: pathlib.Path) -> List[List[str]]:
        """
        Return a list of commands to be executed to set the wallpaper.

        xfconf-query exposes individual configuration properties per
        output and workspace and thus requires multiple set commands.

        :param path: The image to set as the wallpaper
        :return: A list of commands to be executed
        """
        return [['xfconf-query',
                 '-c', 'xfce4-desktop',
                 '-p', prop,
                 '-s', str(path)]
                for prop in self._properties]

    def get_current(self) -> Optional[pathlib.Path]:
        """
        Return the current wallpaper's path from xfconf-query.

        :return: The current wallpaper's path or None if not retrievable
        """
        try:
            prop = (f'/backdrop/screen0/monitor{self._primary_screen}'
                    f'/workspace0/last-image')
            process = subprocess.check_output(['xfconf-query',
                                               '-c', 'xfce4-desktop',
                                               '-p', prop])

            return pathlib.Path(process.strip().decode())
        except subprocess.CalledProcessError:
            logging.exception('Could not retrieve current wallpaper, call to '
                              'xfconf-query returned a non-zero exit status.')
            sys.exit(1)



#!/usr/bin/env python3

"""
Author: Benedikt Vollmerhaus
License: MIT
"""

from pathlib import Path

from setuptools import find_packages, setup


def get_version() -> str:
    """
    Return the current version from the version module.

    :return: The current version
    """
    base_dir = Path(__file__).parent
    version_file = base_dir / 'blurwal/_version.py'

    version = {}
    exec(version_file.read_text(), version)
    return version['__version__']


VERSION = get_version()

#: The project repository
PROJECT_URL = 'https://gitlab.com/BVollmerhaus/blurwal'
#: The current release archive
RELEASE_URL = f'{PROJECT_URL}/-/archive/{VERSION}/blurwal-{VERSION}.tar.gz'

setup(
    name='BlurWal',
    version=VERSION,
    description='Smoothly blurs the wallpaper when windows are opened.',
    long_description=Path('README.md').read_text(),
    long_description_content_type='text/markdown',
    keywords=['blur', 'wallpaper', 'desktop', 'ricing'],
    license='MIT',

    author='Benedikt Vollmerhaus',
    author_email='pypi@vollmerhaus.org',
    url=PROJECT_URL,
    download_url=RELEASE_URL,

    packages=find_packages(),
    entry_points={'console_scripts': ['blurwal=blurwal.__main__:main']},

    python_requires='>=3.6',
    install_requires=['python-xlib'],
    setup_requires=['pytest-runner'],
    tests_require=['pytest', 'pytest-mock', 'pytest-datadir'],

    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: X11 Applications',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3.6',
        'Topic :: Desktop Environment'
    ],
)



"""
X Gon' Give It to Ya.

Author: Benedikt Vollmerhaus
License: MIT
"""

import logging
from typing import Any, List

import Xlib
from Xlib import display, error
from Xlib.display import drawable


class XConnection:
    """A connection to the X server for retrieving properties."""

    def __init__(self) -> None:
        """Open a connection to X and retrieve the root window."""
        self.display = display.Display()
        self.root_win = self.display.screen().root

    def get_primary_output(self) -> str:
        """
        Return the primary output's name via the RandR extension.

        :return: The primary output's name
        """
        resources = self.root_win.xrandr_get_screen_resources()
        primary_output: int = self.root_win.xrandr_get_output_primary().output
        return self.display.xrandr_get_output_info(
            primary_output, resources.config_timestamp).name

    def get_windows_on_current_workspace(self) -> List[drawable.Window]:
        """
        Return a list of open windows on the current workspace.

        :return: A list of open windows on the current workspace
        """
        current_workspace = self._get_current_workspace()
        return [w for w in self._get_open_windows()
                if self._get_workspace_of_window(w) == current_workspace]

    def _get_current_workspace(self) -> int:
        """
        Return the currently focused workspace.

        :return: The currently focused workspace
        """
        return self._get_property('_NET_CURRENT_DESKTOP', self.root_win)[0]

    def _get_open_windows(self) -> List[drawable.Window]:
        """
        Return a list of open windows across all workspaces.

        :return: A list of open windows across all workspaces
        """
        windows = [self.display.create_resource_object('window', w_id) for w_id
                   in self._get_property('_NET_CLIENT_LIST', self.root_win)]
        return [w for w in windows if self._window_is_visible(w)]

    def _window_is_visible(self, window: drawable.Window) -> bool:
        """
        Check whether the given window is visible/not minimized.

        :param window: The window whose visibility to check
        :return: Whether the window is visible
        """
        state_atoms = self._get_property('_NET_WM_STATE', window) or []
        states = [self.display.get_atom_name(a) for a in state_atoms]
        return '_NET_WM_STATE_HIDDEN' not in states

    def _get_workspace_of_window(self, window: drawable.Window) -> int:
        """
        Return the given window's workspace number or -1 if it's
        invalid or missing the required _NET_WM_DESKTOP property.

        :param window: The window whose workspace number to get
        :return: The window's workspace number or -1 if missing
        """
        try:
            workspace = self._get_property('_NET_WM_DESKTOP', window)
            if workspace is not None:
                return workspace[0]

            logging.warning('Window (id: %s) has no workspace.', window.id)
        except error.BadWindow:
            logging.info('Bad window (id: %s)', window.id)

        return -1

    def _get_property(self, name: str, window: drawable.Window) -> Any:
        """
        Retrieve the value of a property on the given window.

        :param name: The name of the property to retrieve
        :param window: The window to retrieve the property from
        :raises error.BadWindow: if the given window is invalid
        :return: The property's value or None if missing
        """
        prop = window.get_full_property(self.display.get_atom(name),
                                        Xlib.X.AnyPropertyType)
        return prop.value if prop is not None else None



"""
Author: Benedikt Vollmerhaus
License: MIT
"""

import abc
import logging
import pathlib
import subprocess
import sys
from abc import abstractmethod
from typing import List, Optional

from blurwal import paths


class Backend(abc.ABC):
    """Abstract class to be implemented by individual backends."""

    def change_to(self, path: pathlib.Path) -> None:
        """
        Set the given image as the wallpaper via the commands returned
        by the backend's :meth:`get_wallpaper_set_cmds` implementation.

        :param path: The image to set as the wallpaper
        :return: None
        """
        logging.debug('Setting wallpaper to: %s', path)
        for cmd in self.get_wallpaper_set_cmds(path):
            subprocess.run(cmd)

    @abstractmethod
    def get_wallpaper_set_cmds(self, path: pathlib.Path) -> List[List[str]]:
        """
        Return a list of commands to be executed to set the wallpaper.

        Executing multiple commands may be required with some backends,
        e.g. to set the wallpaper on individual outputs or workspaces.

        :param path: The image to set as the wallpaper
        :return: A list of commands to be executed
        """

    @abstractmethod
    def get_current(self) -> Optional[pathlib.Path]:
        """
        Return the current wallpaper's path.

        This may be called rather frequently and also while the backend
        is setting the wallpaper, so keep the retrieval inexpensive and
        accommodate for empty or partially written configuration files,
        if applicable to the backend.

        :return: The current wallpaper's path or None if not retrievable
        """

    def restore_original(self) -> None:
        """
        Restore the original wallpaper.

        :return: None
        """
        try:
            original_path = paths.get_original()
            logging.info('Restoring original wallpaper: %s', original_path)
            self.change_to(original_path)
        except FileNotFoundError:
            logging.error('Could not restore original wallpaper, '
                          'please set it manually.')
            sys.exit(1)



"""
Functions for checking the wallpaper's status.

Author: Benedikt Vollmerhaus
License: MIT
"""

import pathlib
import re
from typing import Optional

from blurwal import paths


def changed_externally(current_path: pathlib.Path) -> bool:
    """
    Check whether the wallpaper has been changed externally.

    :param current_path: The current wallpaper's path
    :return: Whether the wallpaper has been changed externally
    """
    return (not is_transition(current_path) and
            current_path.resolve() != paths.get_original().resolve())


def is_transition(current_path: Optional[pathlib.Path]) -> bool:
    """
    Check whether the current wallpaper is a transition frame.

    If the given path is None, i.e. the current wallpaper could not be
    retrieved, it is assumed to be a transition to avoid regeneration.

    :param current_path: The current wallpaper's path
    :return: Whether the current wallpaper is a transition frame
    """
    if current_path is None:
        return True

    if current_path.parent.resolve() != paths.CACHE_DIR.resolve():
        return False

    return re.match(r'frame-\d+', current_path.name) is not None
